import threading

from actfw_core.capture import Frame
from actfw_core.task import Producer
from actfw_jetson.logger import DEFAULT_LOGGER
from PIL import Image


class NVArgusCameraCapture(Producer):  # type: ignore
    """Camera using nvarguscamerasrc plugin.

    Outputs :class:`~actfw_core.capture.Frame` whose content value is PIL.Image.
    """

    def __init__(self, size, fps, logger=DEFAULT_LOGGER):
        """
        Args:
            size (int, int): capture area resolution
            fps (int): framerate
        """
        import gi

        gi.require_version("Gst", "1.0")
        from gi.repository import GObject, Gst

        self._Gst = Gst

        super(NVArgusCameraCapture, self).__init__()

        self._logger = logger
        self._pipeline = self._Gst.Pipeline()

        bus = self._pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", self._on_bus_error)

        # define elements
        nvarguscamerasrc = self._Gst.ElementFactory.make("nvarguscamerasrc")

        capsfilter1 = self._Gst.ElementFactory.make("capsfilter")
        capsfilter1.set_property(
            "caps",
            self._Gst.caps_from_string(
                f"video/x-raw(memory:NVMM),format=NV12,width={size[0]},height={size[1]},framerate={fps}/1"
            ),
        )

        # converts from NV12 into RGBA to use input data of PIL.Image.frombytes.
        nvvidconv = self._Gst.ElementFactory.make("nvvidconv")
        capsfilter2 = self._Gst.ElementFactory.make("capsfilter")
        capsfilter2.set_property("caps", self._Gst.caps_from_string("video/x-raw,format=RGBA"))

        appsink = self._Gst.ElementFactory.make("appsink")
        appsink.set_property("emit-signals", True)

        # add elements
        self._pipeline.add(nvarguscamerasrc)
        self._pipeline.add(capsfilter1)
        self._pipeline.add(nvvidconv)
        self._pipeline.add(capsfilter2)
        self._pipeline.add(appsink)

        # link elements
        nvarguscamerasrc.link(capsfilter1)
        capsfilter1.link(nvvidconv)
        nvvidconv.link(capsfilter2)
        capsfilter2.link(appsink)

        # subscribe to <new-sample> signal
        appsink.connect("new-sample", _appsink_on_new_sample, self)

        self._pipeline.set_state(self._Gst.State.PLAYING)

        self._glib_loop = GObject.MainLoop()
        threading.Thread(target=self._glib_loop.run).start()

    def run(self):
        # _appsink_on_new_sample produces frames
        pass

    def stop(self):
        self._pipeline.set_state(self._Gst.State.NULL)
        self._glib_loop.quit()

    def _on_bus_error(self, bus, msg):
        self._logger.error("on_error():", msg.parse_error())


def _appsink_on_new_sample(sink, slf):
    # Emit 'pull-sample' signal
    # https://lazka.github.io/pgi-docs/GstApp-1.0/classes/AppSink.html#GstApp.AppSink.signals.pull_sample
    sample = sink.emit("pull-sample")

    if isinstance(sample, slf._Gst.Sample):
        im = _extract_buffer(sample)
        frame = Frame(im)
        slf._outlet(frame)
        return slf._Gst.FlowReturn.OK

    return slf._Gst.FlowReturn.ERROR


def _extract_buffer(sample):
    """Extracts Gst.Buffer from Gst.Sample and converts to PIL.Image"""

    buffer = sample.get_buffer()  # Gst.Buffer
    caps_format = sample.get_caps().get_structure(0)  # Gst.Structure
    w, h = caps_format.get_value("width"), caps_format.get_value("height")

    buffer_size = buffer.get_size()
    return Image.frombuffer("RGBA", (w, h), buffer.extract_dup(0, buffer_size))



import threading

from actfw_jetson.logger import DEFAULT_LOGGER
from PIL import Image


class Display:
    """Display using nvoverlaysink plugin"""

    def __init__(self, size, fps, logger=DEFAULT_LOGGER):
        """
        Args:
            size (int, int): display area resolution
            fps (int): framerate
        """
        import gi

        gi.require_version("Gst", "1.0")
        from gi.repository import GObject, Gst

        self._Gst = Gst

        self._logger = logger
        self._pipeline = self._Gst.Pipeline()

        bus = self._pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", Display._on_bus_error)

        # define elements
        self._appsrc = self._Gst.ElementFactory.make("appsrc")
        self._appsrc.set_property("is-live", True)
        self._appsrc.set_property("block", True)

        capsfilter1 = self._Gst.ElementFactory.make("capsfilter")
        capsfilter1.set_property(
            "caps", self._Gst.caps_from_string(f"video/x-raw,format=RGBA,width={size[0]},height={size[1]},framerate={fps}/1")
        )

        nvvidconv = self._Gst.ElementFactory.make("nvvidconv")

        capsfilter2 = self._Gst.ElementFactory.make("capsfilter")
        capsfilter2.set_property("caps", self._Gst.caps_from_string("video/x-raw(memory:NVMM),format=NV12"))

        nvoverlaysink = self._Gst.ElementFactory.make("nvoverlaysink")

        # add elements
        self._pipeline.add(self._appsrc)
        self._pipeline.add(capsfilter1)
        self._pipeline.add(nvvidconv)
        self._pipeline.add(capsfilter2)
        self._pipeline.add(nvoverlaysink)

        # link elements
        self._appsrc.link(capsfilter1)
        capsfilter1.link(nvvidconv)
        nvvidconv.link(capsfilter2)
        capsfilter2.link(nvoverlaysink)

        self._pipeline.set_state(self._Gst.State.PLAYING)

        self._glib_loop = GObject.MainLoop()
        threading.Thread(target=self._glib_loop.run).start()

    def update(self, src_im: Image) -> "Gst.FlowReturn":  # type: ignore  # noqa F821
        """
        Update display.

        Args:
            src_im (PIL.Image): update image (RGB)
        """
        gst_buffer = self._im_to_gst_buffer(src_im)
        self._appsrc.emit("push-buffer", gst_buffer)
        return self._Gst.FlowReturn.OK

    def stop(self):
        self._pipeline.set_state(self._Gst.State.NULL)
        self._glib_loop.quit()

    def _on_bus_error(self, bus: "Gst.Bus", msg: "Gst.Message") -> None:  # type: ignore  # noqa F821
        self._logger.error("on_error():", msg.parse_error())

    def _im_to_gst_buffer(self, im: Image) -> "Gst.Buffer":  # type: ignore  # noqa F821
        """Converts PIL Image (RGB) to Gst.Buffer (RGBA)"""
        im.putalpha(255)
        return self._Gst.Buffer.new_wrapped(im.tobytes())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['actfw_jetson']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=5,<6', 'PyGObject>=3.26.1,<4.0.0', 'actfw-core>=2.1.1,<3.0.0']

setup_kwargs = {
    'name': 'actfw-jetson',
    'version': '0.1.2',
    'description': "actfw's additional components for Jetson series",
    'long_description': '# actfw-jetson\n\nactfw\'s components for Jetson series.\nactfw is a framework for Actcast Application written in Python.\n\n## Installation\n\n```console\nsudo apt-get update\nsudo apt-get install -y python3-pip python3-pil \n\n# Install GStreamer dependencies (some components in actfw-jetson uses GStreamer in implementation)\nsudo apt-get install -y libgstreamer1.0-dev libgirepository1.0-dev ibgstreamer-plugins-base1.0-dev libglib2.0-dev libcairo2-dev\n\npip3 install actfw-jetson\n```\n\n## Document\n\n- [API References](https://idein.github.io/actfw-jetson/latest/)\n\n## Usage\n\nSee [actfw-core](https://github.com/Idein/actfw-core) for basic usage.\n\nSince actfw-jetson uses GStreamer to implement some components, an application using actfw-jetson may have to initialize GStreamer library before using actfw-jetson\'s components.\n\n```python\nif __name__ == \'__main__\':\n    import gi\n    gi.require_version(\'Gst\', \'1.0\')\n    from gi.repository import Gst\n    Gst.init(None)\n\n    main()\n```\n\nactfw-jetson provides:\n\n- `actfw_jetson.Display` : Display using `nvoverlaysink` element in [NVIDIA\'s Accelerated GStreamer](https://docs.nvidia.com/jetson/l4t/index.html#page/Tegra%20Linux%20Driver%20Package%20Development%20Guide/accelerated_gstreamer.html).\n\n## Example\n\n- `example/hello_jetson` : The simplest application example for Jetson\n  - Use HDMI display as 1280x720 area\n  - Generate 1280x720 single-colored image\n  - Draw "Hello, Actcast!" text\n  - Display it as 1280x720 image\n  - Notice message for each frame\n  - Support application heartbeat\n  - Support "Take Photo" command\n  - Depends: fonts-dejavu-core\n\n## Development Guide\n\n### Installation of dev requirements\n\n```console\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python -\npoetry install\n```\n\n### Running tests\n\n```console\npoetry run nose2 -v\n```\n\n### Running examples\n\n#### hello_jetson\n\nDisplays a red rectangle and greeting text on it on HDMI display.\n\nRun on a Jetson Nano connected to HDMI display:\n\n```console\napt-get install fonts-dejavu-core\npoetry run python example/hello_jetson\n```\n\n#### camera_display\n\nDisplays camera input on HDMI display.\n\nRun on a Jetson Nano with CSI camera and HDMI display:\n\n```console\npoetry run python example/camera_display\n```\n\n### Releasing package & API doc\n\nCI will automatically do.\nFollow the following branch/tag rules.\n\n1. Make changes for next version in `master` branch (via pull-requests).\n2. Update `version` field in `pyproject.toml` with new version in `master` branch.\n3. Create Git tag from `master` branch\'s HEAD named `release-<New version>`. E.g. `release-1.4.0`.\n4. Then CI will build/upload package to PyPI & API doc to GitHub Pages.\n',
    'author': 'Idein Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Idein/actfw-jetson',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
